#include "motio.h"
#include <string.h>
#include <math.h>

	struct CQuadrantGrabber
        : Motio
	{
        static uint myabs(int i)
        {
            if ( i<0 ) return (-i);
            return i;
        }
		//const static uint threshold = 100;
		const static uint numQuadrants = 9; // must be odd to have a center value!
		uint *cp; // last buffer
		uint quad[numQuadrants][numQuadrants]; // scaled down matrix
		float qiw, qih; // scale factors, pixel to quad
		float mx, my, mv; // pos x,y and value of maximum
		float dx, dy; // delta pos of max;
		float lx, ly; // last pos
		uint w,h;

		CQuadrantGrabber()
			: cp(0), qiw(0), qih(0), mx(0), my(0), mv(60), dx(0), dy(0), lx(0), ly(0),w(0),h(0)
		{
			memset(quad,0,numQuadrants*numQuadrants*sizeof(uint));
		}

		virtual ~CQuadrantGrabber()
		{
			if(cp) delete[] cp;
		}

		virtual uint processRGB( void * pixel, uint w, uint h, uint threshold )
		{
		    if ( w != this->w || h != this->h )
		    {
		        if ( cp ) delete [] cp;
		        cp = 0;
            }
			if ( ! cp )
			{
				cp = new uint[ w*h ];
				memset( cp, 0, w*h*sizeof(uint) );
				qiw = float(numQuadrants) / w;
				qih = float(numQuadrants) / h;
				this->w = w;
				this->h = h;
			}

			//memset(quad,0,4*4*sizeof(uint));
			for ( uint j=0; j<numQuadrants; j++ )
			{
				for ( uint i=0; i<numQuadrants; i++ )
				{
					quad[ j ][ i ] = 0;
				}
			}

			// sample pixelbuffer for motion:
			struct _rgb { unsigned char b,g,r; };
			_rgb * pix = (_rgb*)pixel;

			uint scan = 0;
			for ( uint j=0; j<h-1; j++ )
			{
				uint qj  = uint(j*qih);
				for ( uint i=0; i<w-1; i++ )
				{
					// sample with a little blur right/down to denoise.
					// can't use up/left, since i'm drawing blue motion there..
					uint p = pix[ i   +scan   ].g
						   + pix[ i   +scan   ].g
						   + pix[ i   +scan+w ].g
						   + pix[ i+1 +scan   ].g
						   + pix[ i+1 +scan+w ].g;

					// motion detected:
					if ( myabs(int(p - cp[i+scan])) > threshold )
					{
						// increase hit quadfield:
						quad[ qj ][ int(i*qiw) ] ++;
						// visualize motion:
						pix[ i+scan ].b = 255;
					}
					// keep mean value for next iter:
					cp[i+scan] = (p + cp[i+scan]) / 2;
				}
				scan += w;
			}

			// search for peak val:
			uint mi=2000,mj=2000;

			mv = mv * 0.90f; // friction or decay
			for ( uint j=0; j<numQuadrants; j++ )
			{
				for ( uint i=0; i<numQuadrants; i++ )
				{
					if ( (uint)mv < quad[ j ][ i ] )
					{
						mi = i;
						mj = j;
						mv = quad[ j ][ i ];
					}
				}
			}
			lx = mx;
			ly = my;
			mx = mi==2000?mx:mi;//(mx+mi)/2;//mi;//
			my = mj==2000?my:mj;//(my+mj)/2; //mj;//
			dx = lx - mx;
			dy = ly - my;
			// visualize quads:
			uint pSize = 8;
			uint pw = numQuadrants * pSize;
			for ( uint j=0; j<pw; j++ )
			{
				for ( uint i=0; i<pw; i++ )
				{
					uint id = (w-10-pw+i)+w*(j+10);
					uint is = i/pSize;
					uint js = j/pSize;
					uint v = quad[ js ][ is ] / 16;
					if ( (is==mx)&&(js==my) )
					{
						pix[id].r = v*8;
						pix[id].g = 0;
						pix[id].b = 0;
					}
					else
					{
						pix[id].r = v;
						pix[id].g = v;
						pix[id].b = v;
					}
				}
			}

			return 1;
		}
		virtual float peakX() const
		{
			return floor(0.5f + mx - numQuadrants*0.5) / (numQuadrants*0.5);
		}
		virtual float peakY() const
		{
			return floor(0.5f + my - numQuadrants*0.5) / (numQuadrants*0.5);
		}
		virtual float peakV() const
		{
			return mv;
		}
		virtual float peakVelX() const
		{
			return dx;
		}
		virtual float peakVelY() const
		{
			return dy;
		}
		virtual float matrix( uint x, uint y )
		{
			return quad[w][h];
		}

	}; // CQuadrantGrabber


	Motio * CreateMotio()
	{
	    return new CQuadrantGrabber();
    }

